/*
	File:		USBTabletModule.c

	Contains:	HID Module for USB Wacom

	Version:	xxx put version here xxx

	Copyright:	 1997-1998 by Apple Computer, Inc., all rights reserved.

*/


#include <Types.h>
#include <Devices.h>
#include <processes.h>
#include <DriverServices.h>
#include <USB.h>

#include "USBTabletModule.h"

usbWacomPBStruct myWacomPB;

void InitParamBlock(USBDeviceRef theDeviceRef, USBPB * paramblock)
{
	paramblock->usbReference = theDeviceRef;
	paramblock->pbVersion = kUSBCurrentPBVersion;
	paramblock->usbStatus = noErr;
	paramblock->usbBuffer = nil;		
	paramblock->usbActCount = 0;
	paramblock->usbReqCount = 0;
	paramblock->usbFlags = 0;
	paramblock->usbOther = 0;
}



Boolean immediateError(OSStatus err)
{
	return((err != kUSBPending) && (err != noErr) );
}

void InitiateTransactionProc(USBPB *pb)
{
register usbWacomPBStruct *pWacomPB;
OSStatus myErr;

	pWacomPB = (usbWacomPBStruct *)(pb);
	pWacomPB->transDepth++;
	if (pWacomPB->transDepth < 0)
	{
		USBExpertFatalError(pWacomPB->deviceRef, kUSBInternalErr, "\pWacom Driver: transDepth < 0 (initiation)", pWacomPB->pb.usbRefcon );
	}
	
	if (pWacomPB->transDepth > 1)
	{
		USBExpertFatalError(pWacomPB->deviceRef, kUSBInternalErr, "\pWacom Driver: transDepth > 1 (initiation)", pWacomPB->pb.usbRefcon );
	}
	
	switch(pWacomPB->pb.usbRefcon & ~kRetryTransaction)
	{
		case kFindInterface:
			InitParamBlock(pWacomPB->deviceRef, &pWacomPB->pb);
			pWacomPB->pb.usb.cntl.WIndex = 0;
			pWacomPB->pb.usb.cntl.WValue = 0;
			pWacomPB->pb.usbClassType = kUSBHIDInterfaceClass;
			pWacomPB->pb.usbSubclass = kUSBBootInterfaceSubClass;
			pWacomPB->pb.usbProtocol = kUSBMouseInterfaceProtocol;
			pWacomPB->pb.usbRefcon |= kCompletionPending;
			pWacomPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBFindNextInterface(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pWacomPB->pb.usbReference, kUSBInternalErr, "\pWacom Driver: USBFindNextInterface (immediate error)", myErr);
				pWacomPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kOpenDevice:
			InitParamBlock(pWacomPB->deviceRef, &pWacomPB->pb);
			pWacomPB->pb.usb.cntl.WValue = pWacomPB->configurationNumber;
			pWacomPB->pb.usbRefcon |= kCompletionPending;
			
			pWacomPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBOpenDevice(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pWacomPB->pb.usbReference, kUSBInternalErr, "\pWacom Driver: USBOpenDevice (immediate error)", myErr);
				pWacomPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kNewInterfaceRef:
			InitParamBlock(pWacomPB->deviceRef, &pWacomPB->pb);
			pWacomPB->pb.usb.cntl.WIndex = pWacomPB->interfaceNumber;
			pWacomPB->pb.usbRefcon |= kCompletionPending;
			
			pWacomPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBNewInterfaceRef(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pWacomPB->pb.usbReference, kUSBInternalErr, "\pWacom Driver: USBNewInterfaceRef (immediate error)", myErr);
				pWacomPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kConfigureInterface:
			InitParamBlock(pWacomPB->interfaceRef, &pWacomPB->pb);
			pWacomPB->pb.usbRefcon |= kCompletionPending;
			
			pWacomPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBConfigureInterface(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pWacomPB->pb.usbReference, kUSBInternalErr, "\pWacom Driver: USBConfigureInterface (immediate error)", myErr);
				pWacomPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kSetProtocol:
			InitParamBlock(pWacomPB->interfaceRef, &pWacomPB->pb);
			
			pWacomPB->pb.usb.cntl.BMRequestType = USBMakeBMRequestType(kUSBOut, kUSBClass, kUSBInterface);			
			pWacomPB->pb.usb.cntl.BRequest = kHIDRqSetProtocol;
			pWacomPB->pb.usb.cntl.WValue = kHIDReportProtocolValue; 
			pWacomPB->pb.usb.cntl.WIndex = pWacomPB->interfaceNumber;
			pWacomPB->pb.usbRefcon |= kCompletionPending;
			
			pWacomPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBDeviceRequest(&pWacomPB->pb);
			if (immediateError(myErr))
			{
				USBExpertFatalError(pWacomPB->deviceRef, kUSBInternalErr, "\pWacom Driver: kHIDRqSetProtocol (immediate error)", myErr);
				pWacomPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kGetReportDescriptor:
			InitParamBlock(pWacomPB->deviceRef, &pWacomPB->pb);
			pWacomPB->pb.usb.cntl.BMRequestType = USBMakeBMRequestType(kUSBIn, kUSBStandard, kUSBInterface);
			pWacomPB->pb.usb.cntl.BRequest = kUSBRqGetDescriptor;
			pWacomPB->pb.usb.cntl.WValue = (0x22 << 8) + 1; 			// get the first report descriptor
			pWacomPB->pb.usb.cntl.WIndex = pWacomPB->interfaceNumber;
			pWacomPB->pb.usbReqCount = 0x66;
			pWacomPB->pb.usbBuffer = pWacomPB->hidReportDescriptor;
			
			pWacomPB->pb.usbRefcon |= kCompletionPending;
			pWacomPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBDeviceRequest(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pWacomPB->pb.usbReference, kUSBInternalErr, "\pWacom Driver: kGetReportDescriptor - immediate error", myErr);
			}
			break;
		
		case kFindInterruptPipe:
			InitParamBlock(pWacomPB->interfaceRef, &pWacomPB->pb);
			pWacomPB->pb.usbFlags = kUSBIn;
			pWacomPB->pb.usbClassType = kUSBInterrupt; 
			
			pWacomPB->pb.usbRefcon |= kCompletionPending;
			pWacomPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBFindNextPipe(&pWacomPB->pb);
			if (immediateError(myErr))
			{
				USBExpertFatalError(pWacomPB->deviceRef, kUSBInternalErr, "\pWacom Driver: kFindInterruptPipe (immediate error)", myErr);
				pWacomPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kReadInterruptPipe:
			InitParamBlock(pWacomPB->pipeRef, &pWacomPB->pb);

			pWacomPB->pb.usbBuffer = (Ptr)pWacomPB->hidReport;
			pWacomPB->pb.usbReqCount = 7;
			pWacomPB->pb.usb.cntl.WIndex = pWacomPB->interfaceNumber;	
			
			pWacomPB->pb.usbRefcon |= kCompletionPending;
			pWacomPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBIntRead(&pWacomPB->pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pWacomPB->deviceRef, kUSBInternalErr, "\pWacom Driver: kReadInterruptPipe (ImmediateError)", myErr);
			}
			break;
			
		default:
			USBExpertFatalError(pWacomPB->deviceRef, kUSBInternalErr, "\pWacom Driver: Transaction initiated with bad refcon value", pWacomPB->pb.usbRefcon);
			pWacomPB->pb.usbRefcon = kUndefined + kReturnFromDriver;
			break;
	}
	
// At this point the control is returned to the system.  If a USB transaction
// has been initiated, then it will call the Complete procs
// (below) to handle the results of the transaction.
}


void TransactionCompletionProc(USBPB *pb)
{
register usbWacomPBStruct *pWacomPB;
unsigned char	* errstring;
USBPipeState 	pipeState;

	pWacomPB = (usbWacomPBStruct *)(pb);
	pWacomPB->transDepth--;
	if (pWacomPB->transDepth < 0)
	{
		USBExpertFatalError(pWacomPB->deviceRef, kUSBInternalErr, "\pWacom Driver: transDepth < 0 (completion)", pWacomPB->pb.usbRefcon );
	}
	
	if (pWacomPB->transDepth > 1)
	{
		USBExpertFatalError(pWacomPB->deviceRef, kUSBInternalErr, "\pWacom Driver: transDepth > 1 (completion)", pWacomPB->pb.usbRefcon );
	}
	
	if(pWacomPB->pb.usbStatus != noErr)														// was there an error?
	{
		switch(pWacomPB->pb.usbRefcon & 0x0fff)												// yes, so show where the error occurred
		{
			case kFindInterface:  				errstring = "\pWacom Driver: Error during kFindInterface"; break;
			case kOpenDevice:					errstring = "\pWacom Driver: Error during kOpenDevice"; break;
			case kNewInterfaceRef:				errstring = "\pWacom Driver: Error during kNewInterfaceRef"; break;
			case kConfigureInterface:  			errstring = "\pWacom Driver: Error during kConfigureInterface"; break;
			case kSetProtocol: 	 				errstring = "\pWacom Driver: Error during kSetProtocol"; break;
			case kGetReportDescriptor: 	 		errstring = "\pWacom Driver: Error during kGetReportDescriptor"; break;
			case kFindInterruptPipe: 	 		errstring = "\pWacom Driver: Error during kFindInterruptPipe"; break;
			case kReadInterruptPipe: 	 		errstring = "\pWacom Driver: Error during kReadInterruptPipe"; break;
			default: 	 						errstring = "\pWacom Driver: Error occurred, but state is unknown"; break;
		};
		USBExpertFatalError(pWacomPB->deviceRef, pWacomPB->pb.usbStatus, errstring, (pWacomPB->pb.usbRefcon & 0x0fff));
		
		pWacomPB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);				// set up to retry the transaction
		pWacomPB->pb.usbRefcon |= kRetryTransaction;
		pWacomPB->retryCount--;
		
		if ((!pWacomPB->retryCount)	|| (pWacomPB->pb.usbStatus == kUSBAbortedError))	// have we exhausted the retries?
		{																				// or received an abort?
			USBExpertStatus(pWacomPB->deviceRef, "\pWacom Driver: Pipe abort or unable to recover from error", pWacomPB->deviceRef);
			pWacomPB->pb.usbRefcon = kReturnFromDriver;									// if so, just exit.
		}
		else																			// if it didn't abort and there's retries left, then...
		{
			if (pWacomPB->pipeRef)														// check if the pipe is open.
			{
				USBGetPipeStatusByReference(pWacomPB->pipeRef, &pipeState);				// yes, so what it's state?
				if (pipeState != kUSBActive)											// if it's not active, try to clear it.  It might be stalled...
				{
					USBExpertStatus(pWacomPB->deviceRef, "\pWacom Driver: Pipe is open and stalled, clearing stall...", pWacomPB->deviceRef);
					USBClearPipeStallByReference(pWacomPB->pipeRef);
				}
			}
		}
	}
	else
	{
		pWacomPB->pb.usbRefcon &= ~kRetryTransaction;
		pWacomPB->retryCount = kWacomRetryCount;
	}

	if (pWacomPB->pb.usbRefcon & kCompletionPending)			 
	{												
		pWacomPB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);
		switch(pWacomPB->pb.usbRefcon)
		{
			case kFindInterface:
				if ((pWacomPB->pb.usbClassType == kUSBHIDInterfaceClass) && 
					(pWacomPB->pb.usbSubclass == kUSBBootInterfaceSubClass) &&
					(pWacomPB->pb.usbProtocol == kUSBMouseInterfaceProtocol))
				{
					pWacomPB->pb.usbRefcon = kOpenDevice;
					pWacomPB->interfaceNumber = pWacomPB->pb.usb.cntl.WIndex;
					pWacomPB->configurationNumber = pWacomPB->pb.usb.cntl.WValue;
				}
				else
				{
					pWacomPB->pb.usbRefcon = kReturnFromDriver;
				}
				break;
				
			case kOpenDevice:
				pWacomPB->pb.usbRefcon = kNewInterfaceRef;
				break;
				
			case kNewInterfaceRef:
				pWacomPB->interfaceRef = pWacomPB->pb.usbReference;
				pWacomPB->pb.usbRefcon = kConfigureInterface;
				break;
			
			case kConfigureInterface:
				pWacomPB->pb.usbRefcon = kGetReportDescriptor;
				break;
			
			case kGetReportDescriptor:
				pWacomPB->pb.usbRefcon = kFindInterruptPipe;
				break;
				
			case kFindInterruptPipe:
				if ((pWacomPB->pb.usbClassType == kUSBInterrupt) && 
					(pWacomPB->pb.usbFlags == kUSBIn))
				{
					pWacomPB->hidReportSize = pWacomPB->pb.usb.cntl.WValue;
					pWacomPB->pipeRef = pWacomPB->pb.usbReference;
					pWacomPB->pb.usbRefcon = kReadInterruptPipe;
				}
				else
				{
					pWacomPB->pb.usbRefcon = kReturnFromDriver;
				}
				break;
				
			case kReadInterruptPipe:
				//DebugStr("\pWacom read completed");
				ProcessInterruptReport(pWacomPB->refcon, pWacomPB->hidReport);
				pWacomPB->pb.usbRefcon = kReadInterruptPipe;
				break;

		}
	}
	if (!(pWacomPB->pb.usbRefcon & kReturnFromDriver))
		InitiateTransactionProc(pb);
}

